#!/usr/bin/perl -w  # cmdWhistle.pl 
# heavily influenced by knockAge.pl for timing and disambiguation

use strict;
$|=1; #for non buffered standard output
# require 'sys/syscall.ph';  # for subsecond timing
use Time::HiRes qw( gettimeofday );  # for subsecond timing

my $option = $ARGV[0] || ""; # simple option handling

my $MAX_TIMEOUT_LENGTH = 4;      # maximum length in seconds of tone pattern
my $LISTEN_TIMEOUT =     2;      # timeout value in seconds between tone

my $MAX_TIME_DEV =       100000; # maximum acceptable deviation between recorded
                                 # pattern values and current time values
my $MAX_TONE_DEV =       2;      # maximum acceptable deviation between recorded 
                                 # pattern values and current tone values

my $MAX_AVG_TONE =       5;      # maximum number of samples to be averaged 

my @queTones =     (); # running average of tones detected
my $prevTone =      0; # the last solid tone detected, used for disambiguation
my $prevInterval =  0; # previous interval of time
my @baseTones =    (); # the currently entered tone sequence
my @baseTimes =    (); # the currently entered temporal values
my %toneHash =     (); # tones, times and commands read from ~/.toneFile
my $toneCount =     0; # the current count of tones entered
my $startTime =    ""; # start of a temporal block
my $currTime =     ""; # current time in the time out loop
my $toneAge =       0; # for tone count synchronization
my $timeOut =       0; # to reset the timer window
# begin subroutines

sub getEpochMicroSeconds {
	my ($seconds, $microseconds) = gettimeofday;
	return($microseconds);
}#getEpochMicroSeconds

sub getEpochSeconds {
	my ($seconds, $microseconds) = gettimeofday;
	return($seconds);
}#getEpochSeconds

sub readTones
{
  # using the other components of the sndpeek output might produce results more easily
  # disambiguated from background noise 
  my(undef, undef, undef, $currentTone ) = split " ", $_[0];

  if( @queTones == $MAX_AVG_TONE )
  {
    my $localDiff = 0;
    # check for a solid tone by computing a simple average over the past five
    # time windows
    # simple time related tonal smoothing, so if your tone wavers
    # just a bit it's still counted as one tone
    for my $chkHistTone ( @queTones )
    {
      if( abs($currentTone - $chkHistTone) > $MAX_TONE_DEV )
      {
        $localDiff =1;
        $prevTone = 0;
      }#if deviation less than threshold
    }#for each tone

    if( $localDiff == 0 )
    {
      # make sure the current tone is different than the previous one, this is to 
      # ensure that long duration tones are not acquired as multiple tone events
      # this step up or down will allow you to whistle continuously and pick up the 
      # steps as discrete tone events
      my $upDev   = $currentTone + $MAX_TONE_DEV;
      my $downDev = $currentTone - $MAX_TONE_DEV;
      if( $prevTone > $upDev || $prevTone < $downDev )
      {
        my $currVal = getEpochMicroSeconds();
        my $diffInterval = abs($prevInterval - $currVal);
        if( $option ){
          print "Tone: $currentTone ## last: [$currVal] curr: [$prevInterval] ";
          print "difference is: $diffInterval\n";
        }
        if( $toneCount == 0 ){ $diffInterval = 0 }
        push @baseTones, $currentTone;
        push @baseTimes, $diffInterval;
        $toneCount++;
        $prevInterval = $currVal;
      }#if deviation in tone

      # now set the previous tone to the current tone so a continuous tone is not
      # acquired as multiple tone events
      $prevTone = $currentTone;

    }#if a solid tone has been found

    # if enough tones to create an average have been added, pop one off
    shift @queTones;

  }#if enough tones to create a useful average

  # always push more tones on the avg
  push @queTones, $currentTone;

}#readTones


# readToneFile reads tone sequences and commands from ~/.toneFile
# format is:  tones _#_ times _#_ command _#_ comments
sub readToneFile 
{
  open(TONEFILE,"$ENV{HOME}/.toneFile") or die "no tone file: $!";

    while(<TONEFILE>){

      if( !/^#/ ){

        my @arrLine = split "_#_";
        $toneHash{ "$arrLine[0] $arrLine[1]" }{ tones }   = $arrLine[0];
        $toneHash{ "$arrLine[0] $arrLine[1]" }{ times }   = $arrLine[1];
        $toneHash{ "$arrLine[0] $arrLine[1]" }{ cmd }     = $arrLine[2];
        $toneHash{ "$arrLine[0] $arrLine[1]" }{ comment } = $arrLine[3];

      }#if not a comment line

    }#for each line in file

  close(TONEFILE);

}#readToneFile

sub compareToneSequences 
{
  my $baseCount = @baseTones;
  my $countMatch = 0;  # record how many tones matched

  for my $toneFromFile ( keys %toneHash )
  {
    my @confTones = split " ", $toneHash{$toneFromFile}{tones};
    my @confTimes = split " ", $toneHash{$toneFromFile}{times};

    my $confCount = @confTones;

    next unless( $confCount == $baseCount );

    # as a learning aid, the matching and non matching portions of the
    # comparison are printed out, so at least you can see what is going 
    # wrong while trying to remember your tone codes
    my $pos =0;
    my $toneMatchFail = 0;
    for( @baseTones )
    {
      my $tonDiff = abs($confTones[$pos] - $baseTones[$pos]);
      my $tonStr = "t $pos b $baseTones[$pos] ".
                   "c $confTones[$pos] \n";

      my $timeDiff = abs($confTimes[$pos] - $baseTimes[$pos]);
      my $timStr = "t $pos b $baseTimes[$pos] ".
                   "c $confTimes[$pos] d $timeDiff\n";

      if( $tonDiff > $MAX_TONE_DEV )
      {
        $toneMatchFail = 1;
        if( $option ){ print "NOTE DISSONANCE $tonStr" }
      }else
      {
        if( $option ){ print "NOTE MATCH $tonStr" }
      }#if tone detected outside of deviation

      # if it's an exact match, increment the matching counter
      if( $timeDiff < $MAX_TIME_DEV ){

        if( $option ){ print "TIME MATCH $timStr" }
        $countMatch++;
      
      # if the tones don't match, move on to the next pattern in the list
      }else{
      
        if( $option ){ print "TIME DISSONANCE $timStr" }
        last;
    
      }# deviation check

      $pos++;
    }# for each tone to check 

    if( $countMatch == $confCount && $toneMatchFail == 0 )
    {
      my $cmd =  $toneHash{$toneFromFile}{ cmd };
      if( $option ){ print "run $cmd\n" }
      $cmd =`$cmd`;
      if( $option ){ print "result $cmd\n" }
      last;

    # otherwise, make the count of matches zero, in order to not reset
    }else
    {   
      $countMatch = 0;
    }

  }#for each tone in tone file
  
  # if the match count is zero, exit and don't reset variables so a longer
  # tone sequence can be entered and checked
  if( $countMatch == 0 ){ return() }
  
  # if a match occured, reset the variables so it won't match another pattern
  $toneCount = 0;
  @baseTones = ();
  @baseTimes = ();

}#compareToneSeqeunces


# begin main program logic
if( $option eq "-c" ){

  print "enter a tone sequence:\n";

  $startTime = getEpochSeconds();  # reset time out start

  while( my $sndPeekOutput = <STDIN> )
  {

    $currTime = getEpochSeconds(); 
 
    # check if there has not been a tone in a while 
    if( $currTime - $startTime > $MAX_TIMEOUT_LENGTH ){
  
      $timeOut = 1;  # exit the loop
  
    }else{
 
      # if a tone has been entered before timeout, reset timers so
      # more tones can be entered
 
      if( $toneCount != $toneAge ){
        $startTime = $currTime;   # reset timer for longer delay
        $toneAge = $toneCount;  # synchronize tone counts
      }# if a new tone came in
  
    }# if timer not reached

    readTones( $sndPeekOutput );

    if( $timeOut == 1 ){ last }
  }#while stdin

  if( @baseTones ){
    print "place the following line in $ENV{HOME}/.toneFile\n\n";
    for( @baseTones ){ print "$_ " }
    print "_#_ ";
    for( @baseTimes ){ print "$_ " }
    print "_#_ (command here) _#_ <comments here>\n\n";
  }#if tones entered

}else
{
  # main code loop to listen for tones and run commands
  readToneFile();
  $startTime = getEpochSeconds();

  while( my $sndPeekOutput = <STDIN> )
  {

    $currTime = getEpochSeconds();

    if( $currTime - $startTime > $LISTEN_TIMEOUT ){

      $toneCount = 0;
      @baseTones = ();
      @baseTimes = ();
      $startTime = $currTime;
      if( $option ){ print "listen timeout - resetting tones \n" }

    }else{

      if( $toneCount != $toneAge ){
        $startTime = $currTime;   # reset timer for longer delay
        $toneAge = $toneCount;    # synchronize tone counts
      }# if a new tone came in

      compareToneSequences();

    }#if not reset timeout

    readTones( $sndPeekOutput );

  }#while stdin

}#if option set


# EOF
